import yaml
import argparse
import re
import json
from pathlib import Path

parser = argparse.ArgumentParser(description="Process some integers.")
parser.add_argument("file", type=Path, help="yaml config")
parser.add_argument("--json", action="store_true", help="output parse as json")
parser.add_argument("--out", type=Path, help="write into file")

args = parser.parse_args()


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Λssεrτiδηs                                                                |
# ╚───────────────────────────────────────────────────────────────────────────╝
def check_for_duplicates(host, props):
    seen = set()

    for v in props:
        if v in seen:
            raise SyntaxError(f"Duplicate key on `{host}`: `{v}`")
        else:
            seen.add(v)


def check_duplicate_hosts(data):
    hosts = set()

    for host in data:
        if host in hosts:
            raise SyntaxError(f"Duplicate host: `{host}`")
        hosts.add(host)


def run_assertions(host, props):
    check_for_duplicates(host, props)


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Pαrsεr                                                                    |
# ╚───────────────────────────────────────────────────────────────────────────╝
def expand_for_loop(props):
    expand = list()

    for loop in props.get("for", []):
        r = loop["range"]

        if isinstance(r, dict):
            start = int(r["from"])
            cease = int(r["to"])
            inc = int(r.get("increment", 1))

            iterations = range(start, cease, inc)
        elif isinstance(r, list):
            iterations = r
        else:
            raise TypeError("Expected list or dict in `for` loop `range`")

        pattern = re.compile(rf'\$\{{{loop["variable"]}\}}')

        for i in iterations:
            expand.append(pattern.sub(str(i), loop["template"]))

    return expand


def get_ssh_props(data):
    props = dict()

    for k, v in data["ssh_props"].items():
        props[k] = v if isinstance(v, list) else [v]

    return props


def parse_props(host, props, hosts):
    parsed = dict()

    # Standard ssh properties
    for k, v in props.get("ssh_props", {}).items():
        parsed[k] = v if isinstance(v, list) else [v]

    # For-loops in config
    for loop in props.get("for", []):
        prop = loop["property"]
        r = loop["range"]

        if isinstance(r, dict):
            iterations = range(r["from"], r["to"])
        elif isinstance(r, list):
            iterations = r
        else:
            raise TypeError("Expected list or dict in `for` loop `range`")

        if not parsed.get(prop):
            parsed[prop] = list()

        pattern = re.compile(rf'\$\{{{loop["variable"]}\}}')

        for i in iterations:
            sub = pattern.sub(str(i), loop["template"])
            parsed[prop].append(sub)

    # Inherit from template
    template = props.get("template")

    if template and template not in hosts:
        raise SyntaxError(f"Template `{template}` required by `{host}` is not declared above `{host}` in the yaml")
    elif template:
        for k, v in hosts[props.get("template")].items():
            if k not in parsed:
                parsed[k] = v

    return parsed


def remove_templates(expanded):
    hosts = dict(expanded)

    for host, props in data.items():
        if host.startswith('__'):
            del hosts[host]

    return hosts


def stringify_booleans(props):
    pattern = re.compile(r"(True|False)$")

    def replace_bools(s):
        m = pattern.search(s)

        if m and m[1] == "True":
            return pattern.sub("true", s)
        elif m and m[1] == "False":
            return pattern.sub("false", s)
        else:
            return s

    return [replace_bools(prop) for prop in props]


def parse_yaml(data):
    hosts = dict()
    check_duplicate_hosts(data)

    for host, props in data.items():
        hosts[host] = parse_props(host, props, hosts)

    hosts = remove_templates(hosts)
    return hosts


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Fδrmαττεr                                                                 |
# ╚───────────────────────────────────────────────────────────────────────────╝
def to_ssh_config_string(parse):
    s = str()

    for host, keys in parse.items():
        s += f"Host {host}\n"
        for key, values in keys.items():
            for value in values:
                if isinstance(value, bool):
                    s += f"\t{key} {'true' if value else 'no'}\n"
                else:
                    s += f"\t{key} {value}\n"
        s += "\n"

    return s


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Mαiη                                                                      |
# ╚───────────────────────────────────────────────────────────────────────────╝
with open(args.file, "r") as f:
    data = yaml.safe_load(f)

parse = parse_yaml(data)

if args.json:
    print(json.dumps(parse))
elif args.out:
    with open(args.out, 'w') as f:
        f.write(to_ssh_config_string(parse))
else:
    print(to_ssh_config_string(parse), end="")
